/* Code specific to x86 hosts running Mac OS X */

// get access to mcontext and thread_state fields with their normal names,
// not with everything prefixed by __ due to namespace cleanliness paranoia
#define __DARWIN_UNIX03 0

#include <string.h>
#include <signal.h>
#include <assert.h>
#include <ucontext.h>

#include <architecture/i386/table.h>
#include <i386/user_ldt.h>

#include "vx32.h"
#include "vx32impl.h"
#include "os.h"


// First LDT selector number to use for our translator segments.
// Avoid stepping on the low selectors used by the kernel.
// XXX is there a way to know how many selectors the kernel uses?
#define SELOFS	((32*8) + 4 + 3)		// 4=LDT, 3=RPL


static void setbase(union ldt_entry *desc, unsigned long base)
{
	desc->code.base00 = base & 0xffff;
	desc->code.base16 = (base >> 16) & 0xff;
	desc->code.base24 = base >> 24;
}

static void setlimit(union ldt_entry *desc, unsigned long limit)
{
	desc->code.limit00 = limit & 0xffff;
	desc->code.limit16 = limit >> 16;
}


// Set up LDT segments needed by the instruction translater
// whenever a vx32 process's memory is mapped.
int vxemu_map(vxemu *emu, vxmmap *mm)
{
	struct vxproc *vxp = emu->proc;

	emu->datasel = SELOFS + vxp->vxpno * 16;
	emu->emusel = emu->datasel + 8;

	if (emu->ldt_base != (uintptr_t)mm->base || emu->ldt_size != mm->size) {

		// Set up the process's data segment selector (for DS,ES,SS).
		union ldt_entry desc;
		memset(&desc, 0, sizeof(desc));
		setbase(&desc, (unsigned long)mm->base);
		setlimit(&desc, (mm->size - 1) >> VXPAGESHIFT);
		desc.data.type = DESC_DATA_WRITE;
		desc.data.dpl = 3;
		desc.data.present = 1;
		desc.data.stksz = DESC_DATA_32B;
		desc.data.granular = 1;
		if (i386_set_ldt(emu->datasel >> 3, &desc, 1) < 0)
			return -1;

		// Set up the process's vxemu segment selector (for FS).
		setbase(&desc, (unsigned long)emu);
		setlimit(&desc, (VXCODEBUFSIZE - 1) >> VXPAGESHIFT);
		if (i386_set_ldt(emu->emusel >> 3, &desc, 1) < 0)
			return -1;

		emu->ldt_base = (uintptr_t)mm->base;
		emu->ldt_size = mm->size;
	}

	return 0;
}

// Note: it's not generally safe to call printf from here,
// even just for debugging, because we're running on a small signal stack.
int vx32_sighandler(int signo, siginfo_t *si, void *v)
{
	// In Darwin, 'mcontext_t' is actually a pointer...
	ucontext_t *uc = v;
	mcontext_t ctx = uc->uc_mcontext;

	// We can't be sure that vxemu is running,
	// and thus that %VSEG is actually mapped to a
	// valid vxemu.  The only way to tell is to look at %VSEG.

	// First sanity check vxproc segment number.
	uint16_t vs;
	asm("movw %"VSEGSTR",%0"
		: "=r" (vs));
	if ((vs & 15) != 15)	// 8 (emu), LDT, RPL=3
		return 0;

	// See if we're in the context of the translator,
	// by testing whether DS is one of our LDT selectors.
	// XXX the process could use other LDT selectors.
	if (ctx->ss.ds < SELOFS) {
		return 0;
	}

	// Okay, assume mapped; check for vxemu.
	uint32_t magic;
	asm("movl %"VSEGSTR":%1,%0"
		: "=r" (magic)
		: "m" (((vxemu*)0)->magic));
	if (magic != VXEMU_MAGIC)
		return 0;

	// Okay, we're convinced.

	// Find current vxproc and vxemu.
	vxproc *vxp;
	asm("mov %"VSEGSTR":%1,%0"
		: "=r" (vxp)
		: "m" (((vxemu*)0)->proc));
	vxemu *emu = vxp->emu;

#ifdef i386
	size_t trapeip = ctx->ss.eip;
#else
	size_t trapeip = ctx->ss.rip;
#endif

	// Get back our regular host segment register state,
	// so that thread-local storage and such works.
	vxrun_cleanup(emu);

	switch(signo){
	case SIGSEGV:
	case SIGBUS:
		emu->cpu.trap = VXTRAP_PAGEFAULT;
		break;
	
	case SIGFPE:
		emu->cpu.trap = VXTRAP_FLOAT;
		break;

	default:
		emu->cpu.trap = VXTRAP_SIGNAL + signo;;
		break;
	}

	if (ctx->ss.ds == vs - 8) {
		// Copy execution state into emu.
		emu->cpu.reg[EAX] = ctx->ss.eax;
		emu->cpu.reg[EBX] = ctx->ss.ebx;
		emu->cpu.reg[ECX] = ctx->ss.ecx;
		emu->cpu.reg[EDX] = ctx->ss.edx;
		emu->cpu.reg[ESI] =  ctx->ss.esi;
		emu->cpu.reg[EDI] = ctx->ss.edi;
		emu->cpu.reg[ESP] = ctx->ss.esp;	// or esp_at_signal ???
		emu->cpu.reg[EBP] = ctx->ss.ebp;
		emu->cpu.eip = 0xffffffff;  // not yet known
		emu->cpu.traperr = ctx->es.err;
		emu->cpu.eflags = ctx->ss.eflags;
		emu->cpu.trapva = ctx->es.faultvaddr;
		trapeip = ctx->ss.eip;
	} else {
		trapeip = 0xffffffff;
		emu->cpu.trap |= VXTRAP_INTERNAL;
		emu->cpu.trapva = ctx->es.faultvaddr;
		emu->cpu.traperr = ctx->es.err;
	}

	vxemu_sighandler(emu, ctx->ss.eip);
	return 0;	// Not reached
}

