// Code specific to x86 hosts running FreeBSD.

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <ucontext.h>
#include <machine/ucontext.h>

#include <machine/segments.h>
#include <machine/sysarch.h>

#include "vx32.h"
#include "vx32impl.h"
#include "os.h"

// First LDT selector number to use for our translator segments.
// Avoid stepping on the low NLDT selectors defined by the kernel.
// XXX I have the impression some versions of FreeBSD TLS
// may also use variable LDT segments; should avoid stepping on them...
#define SELOFS	((32*8) + 4 + 3)	// 4=LDT, 3=RPL


static void setbase(struct segment_descriptor *desc, unsigned long base)
{
	desc->sd_lobase = base & 0xffffff;
	desc->sd_hibase = base >> 24;
}

static void setlimit(struct segment_descriptor *desc, unsigned long limit)
{
	desc->sd_lolimit = limit & 0xffff;
	desc->sd_hilimit = limit >> 16;
}


extern int modify_ldt(int, void*, unsigned long);

int vxemu_map(vxemu *emu, vxmmap *mm)
{
	struct vxproc *vxp;
	union descriptor desc;
	
	vxp = emu->proc;
	emu->datasel = vxp->vxpno * 16 + 32*8 + 4 + 3;	// 4=LDT, 3=RPL
	emu->emusel = emu->datasel + 8;

//fprintf(stderr, "sel data=%x emu=%x\n", emu->datasel, emu->emusel);
	if (emu->ldt_base != (uintptr_t)mm->base || emu->ldt_size != mm->size) {
		// Set up the process's data segment selector (for DS,ES,SS).
		memset(&desc, 0, sizeof(desc));
		setbase(&desc.sd, (unsigned long)mm->base);
		setlimit(&desc.sd, (mm->size - 1) >> VXPAGESHIFT);
		desc.sd.sd_type = SDT_MEMRWA;
		desc.sd.sd_dpl = 3;
		desc.sd.sd_p = 1;
		desc.sd.sd_def32 = 1;
		desc.sd.sd_gran = 1;
		if (i386_set_ldt(emu->datasel >> 3, &desc, 1) < 0)
			return -1;
	
		// Set up the process's vxemu segment selector (for FS).
		setbase(&desc.sd, (unsigned long)emu);
		setlimit(&desc.sd, (VXCODEBUFSIZE - 1) >> VXPAGESHIFT);
		if (i386_set_ldt(emu->emusel >> 3, &desc, 1) < 0)
			return -1;
	
		emu->ldt_base = (uintptr_t)mm->base;
		emu->ldt_size = mm->size;
	}

#ifdef __x86_64
	// Set up 32-bit mode code and data segments (not vxproc-specific),
	// giving access to the full low 32-bit of linear address space.
	// The code segment is necessary to get into 32-bit compatibility mode;
	// the data segment is needed because Linux for x86-64
	// doesn't give 64-bit processes a "real" data segment by default
	// but instead just loads zero into the data segment selectors!
	emu->runptr.sel = FLATCODE;
	desc.entry_number = emu->runptr.sel / 8;
	desc.base_addr = 0;
	desc.limit = 0xfffff;
	desc.contents = MODIFY_LDT_CONTENTS_CODE;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

	desc.entry_number = FLATDATA / 8;
	desc.contents = MODIFY_LDT_CONTENTS_DATA;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

	// Set up a far return vector in emu->retptr
	// for getting back into 64-bit long mode.
	extern void vxrun_return();
	asm volatile("movw %%cs,%0" : "=r" (emu->retptr.sel));
	emu->retptr.ofs = (uint32_t)(intptr_t)vxrun_return;
#endif

	return 0;
}

static void dumpmcontext(mcontext_t *ctx, uint32_t cr2)
{
#ifdef i386
	fprintf(stderr, 
		"eax %08x  ebx %08x ecx %08x  edx %08x\n"
		"esi %08x  edi %08x ebp %08x  esp %08x\n"
		"eip %08x  efl %08x  cs %04x\n"
		"err %08x  trapno %08x  cr2 %08x\n",
		ctx->mc_eax, ctx->mc_ebx, ctx->mc_ecx, ctx->mc_edx,
		ctx->mc_esi, ctx->mc_edi, ctx->mc_ebp, ctx->mc_esp,
		ctx->mc_eip, ctx->mc_eflags, ctx->mc_cs,
		ctx->mc_err, ctx->mc_trapno, cr2);
#else
	fprintf(stderr, 
		"rax %016lx  rbx %016lx\nrcx %016lx  rdx %016lx\n"
		"rsi %016lx  rdi %016lx\nrbp %016lx  rsp %016lx\n"
		"r8  %016lx  r9  %016lx\nr10 %016lx  r11 %016lx\n"
		"r12 %016lx  r13 %016lx\nr14 %016lx  r15 %016lx\n"
		"rip %016lx  efl %016lx  cs %04x  ss %04x\n"
		"err %016lx  trapno %016lx  cr2 %016lx\n",
		ctx->rax, ctx->rbx, ctx->rcx, ctx->rdx,
		ctx->rsi, ctx->rdi, ctx->rbp, ctx->rsp,
		ctx->r8, ctx->r9, ctx->r10, ctx->r11,
		ctx->r12, ctx->r13, ctx->r14, ctx->r15,
		ctx->rip, ctx->eflags, ctx->cs, ctx->__pad0,
		ctx->err, ctx->trapno, ctx->cr2);
#endif
}

int vx32_sighandler(int signo, siginfo_t *si, void *v)
{
	uint32_t trapeip;
	uint32_t magic;
	uint16_t vs;
	vxproc *vxp;
	vxemu *emu;
	ucontext_t *uc;
	mcontext_t *mc;
	uint32_t cr2;

	uc = v;
	mc = &uc->uc_mcontext;

	// same layout, and sigcontext is more convenient...
	cr2 = (uint32_t)si->si_addr;

	// We can't be sure that vxemu is running,
	// and thus that %VSEG is actually mapped to a
	// valid vxemu.  The only way to tell is to look at %VSEG.

	// First sanity check vxproc segment number.
	asm("movw %"VSEGSTR",%0"
		: "=r" (vs));
	if ((vs & 15) != 15)	// 8 (emu), LDT, RPL=3
		return 0;

	// Okay, assume mapped; check for vxemu.
	asm("movl %"VSEGSTR":%1,%0"
		: "=r" (magic)
		: "m" (((vxemu*)0)->magic));
	if (magic != VXEMU_MAGIC)
		return 0;

	// Okay, we're convinced.

	// Find current vxproc and vxemu.
	asm("mov %"VSEGSTR":%1,%0"
		: "=r" (vxp)
		: "m" (((vxemu*)0)->proc));
	emu = vxp->emu;

	// Get back our regular host segment register state,
	// so that thread-local storage and such works.
	vxrun_cleanup(emu);
	
	fprintf(stderr, "siginfo: code %#x status %#x addr %#x\n", si->si_code, si->si_status, (int)si->si_addr);
	
	
	dumpmcontext(mc, cr2);

	switch(signo){
	case SIGSEGV:
	case SIGBUS:
//		abort();
		emu->cpu.trap = VXTRAP_PAGEFAULT;
		break;
	
	case SIGFPE:
		emu->cpu.trap = VXTRAP_FLOAT;
		break;

	default:
		emu->cpu.trap = VXTRAP_SIGNAL + signo;;
		break;
	}

	if (mc->mc_ds == vs - 8) {
		// Copy execution state into emu.
		emu->cpu.reg[EAX] = mc->mc_eax;
		emu->cpu.reg[EBX] = mc->mc_ebx;
		emu->cpu.reg[ECX] = mc->mc_ecx;
		emu->cpu.reg[EDX] = mc->mc_edx;
		emu->cpu.reg[ESI] =  mc->mc_esi;
		emu->cpu.reg[EDI] = mc->mc_edi;
		emu->cpu.reg[ESP] = mc->mc_esp;	// or esp_at_signal ???
		emu->cpu.reg[EBP] = mc->mc_ebp;
		emu->cpu.eip = 0xffffffff;  // not yet known
		emu->cpu.eflags = mc->mc_eflags;
		emu->cpu.trapva = cr2;
		trapeip = mc->mc_eip;
	} else {
		trapeip = 0xffffffff;
		emu->cpu.trap |= VXTRAP_INTERNAL;
		emu->cpu.trapva = cr2;
	}
	emu->cpu.traperr = 0xffffffff;  // broken until freebsd 7

	vxemu_sighandler(emu, mc->mc_eip);
	return 0;	// Not reached
}

